/*
 * This program is a fast hack to reflash a 
 * half dead Linksys WVC54G with a living
 * rrload boot loader. Switch on while pressing
 * the reset button. If the pwr led blinks you
 * may have a chance to flash. Also: Get the
 * libpcap sources into the parent dir for
 * include files.
 *
 * Copyleft 05/2005 sven-ola@gmx.de
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include <pcap.h>
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include "../libpcap-0.8.3/pcap-int.h"

#define lengthof(arr) (sizeof(arr) / sizeof(arr[0]))
#define offsetof(type,member) ((size_t) &((type*)0)->member)

#define FLASH_SIZE 0x00400000

// Defines from rrload/Download.c

#define  KFS_START     0x00010000
#define  KFS_END       (FLASH_SIZE-(64<<10))

#define  WRITE_EALL             1
#define  WRITE_UPGRADE          2
#define  ERASE_EALL	        3
#define  ERASE_UPGRADE	        4
#define  ERASE_VERIFY_EALL      5
#define  ERASE_VERIFY_UPGRADE   6

#define VCI_LEN   56

typedef struct {
	unsigned char	DestAddress[6];
	unsigned char	SrcAddress[6];
	unsigned short	sap;
	unsigned short	wcmd;
	unsigned short	wsequence;
	unsigned short	woffset;
	unsigned short	wsegment;
	unsigned short	wLen;
	unsigned char	bData[600];
} DLC __attribute__((packed));

#define GET_VERSION_INFO	0x0000
#define DOWN_REQUEST		0x0001
#define DOWN_DATA		0x0002
#define DOWN_RESET		0x0003
#define DOWN_VERIFY		0x0004
#define DOWN_EALL		0x0005
#define DOWN_REPLY		0x0006
#define AS_NORMAL		0x0008
#define AS_FORCE		0x0009

#define RRLOAD_SAP		0x8888

int get_version_info(pcap_t *pd, unsigned char* mac_src, unsigned char* mac_dst)
// Get the device version ID (therefore check connection)
{
	DLC dlc;
	bzero(&dlc, sizeof(dlc));
	memcpy(dlc.DestAddress, mac_dst, sizeof(dlc.DestAddress));
	memcpy(dlc.SrcAddress, mac_src, sizeof(dlc.SrcAddress));
	dlc.sap = RRLOAD_SAP;
	dlc.wcmd = GET_VERSION_INFO;
	
	if ((int)offsetof(DLC, bData) + dlc.wLen != write(pd->fd, &dlc, offsetof(DLC, bData) + dlc.wLen))
	{
		fprintf(stderr, "write error to pcap device\n");
		return 0;
	}
	
	struct pcap_pkthdr hdr;
	const unsigned char* packet;
	if (0 == (packet = pcap_next(pd, &hdr))) {
		fprintf(stderr, "no answer packet received\n");
		return 0;
	}
	if (0x20 > hdr.len || 0x00567004 != *((unsigned int*)&packet[0x1c])) {
		fprintf(stderr, "not a wvc54g camera\n");
		return 0;
	}
	
	printf("WVC54G camera found.\n");
	return 1;
}

int down_request(pcap_t *pd, unsigned char* mac_src, unsigned char* mac_dst)
// Request an upgrade (upload all but the loader itself)
{
	DLC dlc;
	bzero(&dlc, sizeof(dlc));
	memcpy(dlc.DestAddress, mac_dst, sizeof(dlc.DestAddress));
	memcpy(dlc.SrcAddress, mac_src, sizeof(dlc.SrcAddress));
	dlc.sap = RRLOAD_SAP;
	dlc.wcmd = DOWN_REQUEST;
	
	printf("Sending update req from %02x:%02x:%02x:%02x:%02x:%02x to %02x:%02x:%02x:%02x:%02x:%02x\n",
	  mac_src[0], mac_src[1], mac_src[2], mac_src[3], mac_src[4], mac_src[5],
	  mac_dst[0], mac_dst[1], mac_dst[2], mac_dst[3], mac_dst[4], mac_dst[5]);
	if ((int)offsetof(DLC, bData) + dlc.wLen != write(pd->fd, &dlc, offsetof(DLC, bData) + dlc.wLen))
	{
		fprintf(stderr, "write error to pcap device\n");
		return 0;
	}
	
	struct pcap_pkthdr hdr;
	DLC* packet;
	if (0 == (packet = (DLC*)pcap_next(pd, &hdr))) {
		fprintf(stderr, "no answer packet received\n");
		return 0;
	}
	if ((int)offsetof(DLC, bData) > hdr.len || 2 != packet->wLen) {
		fprintf(stderr, "unexpected answer, hdr.len=%d, packet.wLen=%d\n", hdr.len, packet->wLen);
		return 0;
	}
	
	printf("Switched to upgrade mode.\n");
	return 1;
}

int down_data(pcap_t *pd, unsigned char* mac_src, unsigned char* mac_dst, 
	unsigned char* buf, ssize_t cnt, size_t ofs, unsigned short seq)
// Upload some data for flashing
{
	DLC dlc;
	bzero(&dlc, sizeof(dlc));
	memcpy(dlc.DestAddress, mac_dst, sizeof(dlc.DestAddress));
	memcpy(dlc.SrcAddress, mac_src, sizeof(dlc.SrcAddress));
	dlc.sap = RRLOAD_SAP;
	dlc.wcmd = DOWN_DATA;
	dlc.wsequence = seq;
	dlc.wLen = cnt;
	dlc.wsegment = ofs >> 4;
	dlc.woffset = ofs & 0x000f;
	memcpy(dlc.bData, buf, cnt);
	
	if ((int)offsetof(DLC, bData) + dlc.wLen != write(pd->fd, &dlc, offsetof(DLC, bData) + dlc.wLen))
	{
		fprintf(stderr, "write error to pcap device\n");
		return 0;
	}
	
	struct pcap_pkthdr hdr;
	DLC* packet;
	if (0 == (packet = (DLC*)pcap_next(pd, &hdr))) {
		fprintf(stderr, "no answer packet received\n");
		return 0;
	}
	if ((int)offsetof(DLC, bData) > hdr.len || 2 != packet->wLen ) {
		// Seems to have sometimes receive probs. Retry.
		//fprintf(stderr, "unexpected answer, hdr.len=%d, packet.wLen=%d\n", hdr.len, packet->wLen);
		if (0 == (packet = (DLC*)pcap_next(pd, &hdr))) {
			fprintf(stderr, "no answer packet received\n");
			return 0;
		}
		if ((int)offsetof(DLC, bData) > hdr.len || 2 != packet->wLen ) {
			fprintf(stderr, "unexpected answer, hdr.len=%d, packet.wLen=%d\n", hdr.len, packet->wLen);
			return 0;
		}
	}
	if (0 != *((unsigned short*)packet->bData)) {
		fprintf(stderr, "Error anwser %d\n", *((unsigned short*)packet->bData));
		return 0;
	}
	if (packet->wsequence != seq) {
		fprintf(stderr, "Out of sequence (%d != %d): Error\n", packet->wsequence, seq);
		return 0;
	}
	
	return 1;
}

int down_reset(pcap_t *pd, unsigned char* mac_src, unsigned char* mac_dst)
// Request a reset
{
	DLC dlc;
	bzero(&dlc, sizeof(dlc));
	memcpy(dlc.DestAddress, mac_dst, sizeof(dlc.DestAddress));
	memcpy(dlc.SrcAddress, mac_src, sizeof(dlc.SrcAddress));
	dlc.sap = RRLOAD_SAP;
	dlc.wcmd = DOWN_RESET;
	
	if ((int)offsetof(DLC, bData) + dlc.wLen != write(pd->fd, &dlc, offsetof(DLC, bData) + dlc.wLen))
	{
		fprintf(stderr, "write error to pcap device\n");
		return 0;
	}
	
	struct pcap_pkthdr hdr;
	DLC* packet;
	if (0 == (packet = (DLC*)pcap_next(pd, &hdr))) {
		fprintf(stderr, "no answer packet received\n");
		return 0;
	}
	if ((int)offsetof(DLC, bData) > hdr.len || 2 != packet->wLen) {
		fprintf(stderr, "unexpected answer, hdr.len=%d, packet.wLen=%d\n", hdr.len, packet->wLen);
		return 0;
	}
	
	printf("Switched to upgrade mode.\n");
	return 1;
}

pcap_t* open_pcap(char* device)
{
	pcap_t *pd = 0;			// packet capture struct pointer
	int snaplen = 200;		// amount of data to capture
	
	char errbuf[PCAP_ERRBUF_SIZE];

	if (0 == device) {
		if (0 == (device = pcap_lookupdev(errbuf))) {
			fprintf(stderr, "pcap_lookupdev: %s\n", errbuf);
			return 0;
		}
	}
	// hardcode: promisc=0, to_ms=500
	if (0 == (pd = pcap_open_live(device, snaplen, 0, 500, errbuf))) {
		fprintf(stderr, "pcap_open_live: %s\n", errbuf);
		return 0;
	}
	
	return pd;
}

static void usage() {
	fprintf(stderr, "Usage: flashwvc -i device -d macaddr flashfile\n");
	exit(1);
}

int main(int argc, char *argv[])
{
	int c;
	char *device = 0;
	unsigned char mac_cmp[ETH_ALEN] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
	unsigned char mac_src[ETH_ALEN] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
	unsigned char mac_dst[ETH_ALEN] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
	while ((c = getopt(argc, argv, "s:d:i:h")) != -1) {
		switch (c) {
			case 's':
			case 'd':
			{
				unsigned int mac[ETH_ALEN];
				sscanf(optarg, "%02x:%02x:%02x:%02x:%02x:%02x", 
					&mac[0], &mac[1], &mac[2],
					&mac[3], &mac[4], &mac[5]);
				int i;
				for(i = 0; i < (int)lengthof(mac); i++) {
					if ('s' == c) {
						mac_src[i] = mac[i] & 0xff;
					}
					else {
						mac_dst[i] = mac[i] & 0xff;
					}
				}
			}
			break;
			case 'i':
				device = optarg; // pcap device
			break;
			default:
				usage();
			break;
		} // switch
	}

	if (optind >= argc || 0 == device ||
		0 == memcmp(mac_cmp, mac_src, lengthof(mac_cmp)) ||
		0 == memcmp(mac_cmp, mac_dst, lengthof(mac_cmp)))
	{
		usage();
	}
	pcap_t* pd = 0;
	if (0 == (pd = open_pcap(device))) exit(1);
	if (0 == get_version_info(pd, mac_src, mac_dst)) exit(1);
	
	int fd = open(argv[optind], O_RDONLY);
	if (-1 == fd) {
		fprintf(stderr, "Error #%d opening %s\n", errno, argv[optind]);
		exit(1);
	}
	if (FLASH_SIZE != lseek(fd, 0, SEEK_END)) {
		fprintf(stderr, "Expeced firmware file of length %d\n", FLASH_SIZE);
		exit(1);
	}
	lseek(fd, 0, SEEK_SET);
	unsigned char buffer[512];
	ssize_t count;
	unsigned short wsequence = 1;
	size_t offset = 0;
	
	if (0 == down_request(pd, mac_src, mac_dst)) exit(1);
	while(0 != (count = read(fd, buffer, sizeof(buffer)))) {
		if (0 == offset) {
			printf("Erasing flash...\n");
		}
		if (0 == down_data(pd, mac_src, mac_dst, buffer, count, offset, wsequence)) exit(1);
		printf("Upload block %d:\r", wsequence);
		wsequence = (wsequence + 1) & 0xffff;
		offset += count;
	}
	printf("\n");
	close(fd);
	printf("Rebooting device...\n");
	if (0 == down_reset(pd, mac_src, mac_dst)) exit(1);
	exit(0);
}
